#ifndef _CU_WQF_ALGEBRA_H
#define _CU_WQF_ALGEBRA_H

#define  thd_t  (threadIdx.x % _NBT)
#define  thd_z  (threadIdx.x /(_NBT) % _NBZ)
#define  thd_y  (threadIdx.x /(_NBT*_NBZ) % _NBY)
#define  thd_x  (threadIdx.x /(_NBT*_NBZ*_NBY) % _NBX)
#define  blk_t  (blockIdx.x % _NGT)
#define  blk_z  (blockIdx.x /(_NGT) % _NGZ)
#define  blk_y  (blockIdx.x /(_NGT*_NGZ) % _NGY)
#define  blk_x  (blockIdx.x /(_NGT*_NGZ*_NGY) % _NGX)

const unsigned int maxThreads = 256;

unsigned int nextPow2( unsigned int x ) {
    --x;
    x |= x >> 1;
    x |= x >> 2;
    x |= x >> 4;
    x |= x >> 8;
    x |= x >> 16;
    return ++x;
}


static __global__ void 
cureverse_eo_kernel(const int *ieo,int *ioe){
  *ioe = (*ieo+1)%2;
}

static __global__ void 
clear_kernel(cuhqfield *y){
#pragma unroll 6
  for(unsigned int i=0;i<CSH;i++) {
    (*y).b[blockIdx.x].yri[i].c[threadIdx.x].x = 0.0f;
    (*y).b[blockIdx.x].yri[i].c[threadIdx.x].y = 0.0f;
    (*y).b[blockIdx.x].yri[i].c[threadIdx.x].z = 0.0f;
    (*y).b[blockIdx.x].yri[i].c[threadIdx.x].w = 0.0f;
  }
}

static __global__ void 
accum_sub_kernel(cuhqfield *a, const cuhqfield *b){
//
// a = a - b
//
  __align__(16) float4 ari,bri;
#pragma unroll 6
  for(unsigned int i=0;i<CSH;i++) {
    ari = (*a).b[blockIdx.x].yri[i].c[threadIdx.x];
    bri = (*b).b[blockIdx.x].yri[i].c[threadIdx.x];
    ari.x = ari.x - bri.x;
    ari.y = ari.y - bri.y;
    ari.z = ari.z - bri.z;
    ari.w = ari.w - bri.w;
    (*a).b[blockIdx.x].yri[i].c[threadIdx.x] = ari;
  }
}

static __global__ void 
accum_mult_kernel(cuhqfield *y, const float *rcoef){
//
// y = rcoef * y;
//
  __align__(16) float4 yri;
#pragma unroll 6
  for(unsigned int i=0;i<CSH;i++) {
    yri = (*y).b[blockIdx.x].yri[i].c[threadIdx.x];
    yri.x = yri.x * (*rcoef);
    yri.y = yri.y * (*rcoef);
    yri.z = yri.z * (*rcoef);
    yri.w = yri.w * (*rcoef);
    (*y).b[blockIdx.x].yri[i].c[threadIdx.x] = yri;
  }
}

static __global__ void 
accum_mult_add_kernel(cuhqfield *w, const float *rcoef, const cuhqfield *y){
//
// w = rcoef * w + y;
//
  __align__(16) float4 yri,wri;
#pragma unroll 6
  for(unsigned int i=0;i<CSH;i++) {
    yri = (*y).b[blockIdx.x].yri[i].c[threadIdx.x];
    wri = (*w).b[blockIdx.x].yri[i].c[threadIdx.x];
    yri.x += wri.x * (*rcoef);
    yri.y += wri.y * (*rcoef);
    yri.z += wri.z * (*rcoef);
    yri.w += wri.w * (*rcoef);
    (*w).b[blockIdx.x].yri[i].c[threadIdx.x] = yri;
  }
}

static __global__ void 
accum_nmult_add_kernel(cuhqfield *w, const float *rcoef, const cuhqfield *y){
//
// w = -rcoef * w + y;
//
  __align__(16) float4 yri,wri;
  __align__(16) float  nrcoef;
  nrcoef = -(*rcoef);
#pragma unroll 6
  for(unsigned int i=0;i<CSH;i++) {
    yri = (*y).b[blockIdx.x].yri[i].c[threadIdx.x];
    wri = (*w).b[blockIdx.x].yri[i].c[threadIdx.x];
    yri.x += wri.x * (nrcoef);
    yri.y += wri.y * (nrcoef);
    yri.z += wri.z * (nrcoef);
    yri.w += wri.w * (nrcoef);
    (*w).b[blockIdx.x].yri[i].c[threadIdx.x] = yri;
  }
}

static __global__ void 
accum_mult_add_kernel(cuhqfield *w, const cuComplex *ccoef, const cuhqfield *y){
//
// w = ccoef * w + y;
//
  __align__(16) float cr,ci,cj;
  __align__(16) float4 wr,wi,yr,yi;
  cr = (*ccoef).x;
  ci = (*ccoef).y;
  cj =-(*ccoef).y;
#pragma unroll 3
  for(unsigned int i=0;i<COL;i++) {
    yr = (*y).b[blockIdx.x].yri[i    ].c[threadIdx.x];
    yi = (*y).b[blockIdx.x].yri[i+COL].c[threadIdx.x];
    wr = (*w).b[blockIdx.x].yri[i    ].c[threadIdx.x];
    wi = (*w).b[blockIdx.x].yri[i+COL].c[threadIdx.x];
    yr.x += cr * wr.x;
    yr.x += cj * wi.x;
    yr.y += cr * wr.y;
    yr.y += cj * wi.y;
    yr.z += cr * wr.z;
    yr.z += cj * wi.z;
    yr.w += cr * wr.w;
    yr.w += cj * wi.w;
    yi.x += cr * wi.x;
    yi.x += ci * wr.x;
    yi.y += cr * wi.y;
    yi.y += ci * wr.y;
    yi.z += cr * wi.z;
    yi.z += ci * wr.z;
    yi.w += cr * wi.w;
    yi.w += ci * wr.w;
    (*w).b[blockIdx.x].yri[i    ].c[threadIdx.x] = yr;
    (*w).b[blockIdx.x].yri[i+COL].c[threadIdx.x] = yi;
  }
}

static __global__ void 
accum_nmult_add_kernel(cuhqfield *w, const cuComplex *ccoef, const cuhqfield *y){
//
// w = -ccoef * w + y;
//
  __align__(16) float cr,ci,cj;
  __align__(16) float4 wr,wi,yr,yi;
  cr =-(*ccoef).x;
  ci =-(*ccoef).y;
  cj =+(*ccoef).y;
#pragma unroll 3
  for(unsigned int i=0;i<COL;i++) {
    yr = (*y).b[blockIdx.x].yri[i    ].c[threadIdx.x];
    yi = (*y).b[blockIdx.x].yri[i+COL].c[threadIdx.x];
    wr = (*w).b[blockIdx.x].yri[i    ].c[threadIdx.x];
    wi = (*w).b[blockIdx.x].yri[i+COL].c[threadIdx.x];
    yr.x += cr * wr.x;
    yr.x += cj * wi.x;
    yr.y += cr * wr.y;
    yr.y += cj * wi.y;
    yr.z += cr * wr.z;
    yr.z += cj * wi.z;
    yr.w += cr * wr.w;
    yr.w += cj * wi.w;
    yi.x += cr * wi.x;
    yi.x += ci * wr.x;
    yi.y += cr * wi.y;
    yi.y += ci * wr.y;
    yi.z += cr * wi.z;
    yi.z += ci * wr.z;
    yi.w += cr * wi.w;
    yi.w += ci * wr.w;
    (*w).b[blockIdx.x].yri[i    ].c[threadIdx.x] = yr;
    (*w).b[blockIdx.x].yri[i+COL].c[threadIdx.x] = yi;
  }
}

static __global__ void 
accum_add_mult_kernel(cuhqfield *w, const cuComplex *ccoef, const cuhqfield *y){
//
// w = w + ccoef * y;
//
  __align__(16) float cr,ci,cj;
  __align__(16) float4 wr,wi,yr,yi;
  cr = (*ccoef).x;
  ci = (*ccoef).y;
  cj =-(*ccoef).y;
#pragma unroll 3
  for(unsigned int i=0;i<COL;i++) {
    yr = (*y).b[blockIdx.x].yri[i    ].c[threadIdx.x];
    yi = (*y).b[blockIdx.x].yri[i+COL].c[threadIdx.x];
    wr = (*w).b[blockIdx.x].yri[i    ].c[threadIdx.x];
    wi = (*w).b[blockIdx.x].yri[i+COL].c[threadIdx.x];
    wr.x += cr * yr.x;
    wr.x += cj * yi.x;
    wr.y += cr * yr.y;
    wr.y += cj * yi.y;
    wr.z += cr * yr.z;
    wr.z += cj * yi.z;
    wr.w += cr * yr.w;
    wr.w += cj * yi.w;
    wi.x += cr * yi.x;
    wi.x += ci * yr.x;
    wi.y += cr * yi.y;
    wi.y += ci * yr.y;
    wi.z += cr * yi.z;
    wi.z += ci * yr.z;
    wi.w += cr * yi.w;
    wi.w += ci * yr.w;
    (*w).b[blockIdx.x].yri[i    ].c[threadIdx.x] = wr;
    (*w).b[blockIdx.x].yri[i+COL].c[threadIdx.x] = wi;
  }
}

static __global__ void 
accum_add_mult_kernel(cuhqfield *w, const float *rcoef, const cuhqfield *y){
//
// w = w + rcoef * y;
//
  __align__(16) float  rr;
  __align__(16) float4 wr,wi,yr,yi;
  rr = *rcoef;
#pragma unroll 3
  for(unsigned int i=0;i<COL;i++) {
    yr = (*y).b[blockIdx.x].yri[i    ].c[threadIdx.x];
    yi = (*y).b[blockIdx.x].yri[i+COL].c[threadIdx.x];
    wr = (*w).b[blockIdx.x].yri[i    ].c[threadIdx.x];
    wi = (*w).b[blockIdx.x].yri[i+COL].c[threadIdx.x];
    wr.x += rr * yr.x;
    wr.y += rr * yr.y;
    wr.z += rr * yr.z;
    wr.w += rr * yr.w;
    wi.x += rr * yi.x;
    wi.y += rr * yi.y;
    wi.z += rr * yi.z;
    wi.w += rr * yi.w;
    (*w).b[blockIdx.x].yri[i    ].c[threadIdx.x] = wr;
    (*w).b[blockIdx.x].yri[i+COL].c[threadIdx.x] = wi;
  }
}

static __global__ void 
accum_sub_mult_kernel(cuhqfield *w, const cuComplex *ccoef, const cuhqfield *y){
//
// w = w - ccoef * y;
//
  __align__(16) float cr,ci,cj;
  __align__(16) float4 wr,wi,yr,yi;
  cr =-(*ccoef).x;
  ci =-(*ccoef).y;
  cj =+(*ccoef).y;
#pragma unroll 3
  for(unsigned int i=0;i<COL;i++) {
    yr = (*y).b[blockIdx.x].yri[i    ].c[threadIdx.x];
    yi = (*y).b[blockIdx.x].yri[i+COL].c[threadIdx.x];
    wr = (*w).b[blockIdx.x].yri[i    ].c[threadIdx.x];
    wi = (*w).b[blockIdx.x].yri[i+COL].c[threadIdx.x];
    wr.x += cr * yr.x;
    wr.x += cj * yi.x;
    wr.y += cr * yr.y;
    wr.y += cj * yi.y;
    wr.z += cr * yr.z;
    wr.z += cj * yi.z;
    wr.w += cr * yr.w;
    wr.w += cj * yi.w;
    wi.x += cr * yi.x;
    wi.x += ci * yr.x;
    wi.y += cr * yi.y;
    wi.y += ci * yr.y;
    wi.z += cr * yi.z;
    wi.z += ci * yr.z;
    wi.w += cr * yi.w;
    wi.w += ci * yr.w;
    (*w).b[blockIdx.x].yri[i    ].c[threadIdx.x] = wr;
    (*w).b[blockIdx.x].yri[i+COL].c[threadIdx.x] = wi;
  }
}

static __global__ void 
accum_sub_mult_kernel(cuhqfield *w, const float *rcoef, const cuhqfield *y){
//
// w = w - rcoef * y;
//
  __align__(16) float rr;
  __align__(16) float4 wr,wi,yr,yi;
  rr =-(*rcoef);
#pragma unroll 3
  for(unsigned int i=0;i<COL;i++) {
    yr = (*y).b[blockIdx.x].yri[i    ].c[threadIdx.x];
    yi = (*y).b[blockIdx.x].yri[i+COL].c[threadIdx.x];
    wr = (*w).b[blockIdx.x].yri[i    ].c[threadIdx.x];
    wi = (*w).b[blockIdx.x].yri[i+COL].c[threadIdx.x];
    wr.x += rr * yr.x;
    wr.y += rr * yr.y;
    wr.z += rr * yr.z;
    wr.w += rr * yr.w;
    wi.x += rr * yi.x;
    wi.y += rr * yi.y;
    wi.z += rr * yi.z;
    wi.w += rr * yi.w;
    (*w).b[blockIdx.x].yri[i    ].c[threadIdx.x] = wr;
    (*w).b[blockIdx.x].yri[i+COL].c[threadIdx.x] = wi;
  }
}

static __global__ void 
prod_kernel(const cuhqfield *y, const cuhqfield *w, cuComplex *gri, const int *ieo){
//
//  gri = <y|w>
//
  float ywr = 0.0f;
  float yxr = 0.0f;
  float ywi = 0.0f;
  float yxi = 0.0f;
#pragma unroll 3
  for(unsigned int i=0;i<COL;i++) {
    __align__(16) float4 wr,wi,yr,yi;
    yr = (*y).b[blockIdx.x].yri[i    ].c[threadIdx.x];
    yi = (*y).b[blockIdx.x].yri[i+COL].c[threadIdx.x];
    wr = (*w).b[blockIdx.x].yri[i    ].c[threadIdx.x];
    wi = (*w).b[blockIdx.x].yri[i+COL].c[threadIdx.x];
    ywr += yr.x * wr.x;
    ywr += yr.y * wr.y;
    ywr += yr.z * wr.z;
    ywr += yr.w * wr.w;
    yxr += yi.x * wi.x;
    yxr += yi.y * wi.y;
    yxr += yi.z * wi.z;
    yxr += yi.w * wi.w;
    ywi += yr.x * wi.x;
    ywi += yr.y * wi.y;
    ywi += yr.z * wi.z;
    ywi += yr.w * wi.w;
    yxi += yi.x * wr.x;
    yxi += yi.y * wr.y;
    yxi += yi.z * wr.z;
    yxi += yi.w * wr.w;
  }
  ywr = ywr + yxr;
  ywi = ywi - yxi;
#ifdef _SF
  {
    int itb    = thd_t + blk_t*_NBT;
    int iz     = thd_z + blk_z*_NBZ;
    int iy     = thd_y + blk_y*_NBY;
    int ix     = thd_x + blk_x*_NBX;
    int ieoxyz = ((*ieo)+(iz+1)+(iy+1)+(ix+1))%2;
    float flag = (float)(1-ieoxyz*((_NTH - itb)/_NTH));
    ywr = ywr * flag;
    ywi = ywi * flag;
  }
#endif

  {
    __shared__ __align__(16) float sywr[CEIL],sywi[CEIL];
#if THREADS != CEIL
    sywr[CEIL - THREADS + threadIdx.x] = 0.0f;
    sywi[CEIL - THREADS + threadIdx.x] = 0.0f;
#endif
    __syncthreads();

    sywr[threadIdx.x] = ywr;
    sywi[threadIdx.x] = ywi;

    __syncthreads();

    for (unsigned int s = CEIL/2; s > 32 ; s >>= 1){
      if (threadIdx.x < s) {
        sywr[threadIdx.x] += sywr[threadIdx.x^s];
        sywi[threadIdx.x] += sywi[threadIdx.x^s];
      }
      __syncthreads();
    }
    if (threadIdx.x < 32) {
      sywr[threadIdx.x] += sywr[threadIdx.x^32];
      sywi[threadIdx.x] += sywi[threadIdx.x^32];
      sywr[threadIdx.x] += sywr[threadIdx.x^16];
      sywi[threadIdx.x] += sywi[threadIdx.x^16];
      sywr[threadIdx.x] += sywr[threadIdx.x^8 ];
      sywi[threadIdx.x] += sywi[threadIdx.x^8 ];
      sywr[threadIdx.x] += sywr[threadIdx.x^4 ];
      sywi[threadIdx.x] += sywi[threadIdx.x^4 ];
      sywr[threadIdx.x] += sywr[threadIdx.x^2 ];
      sywi[threadIdx.x] += sywi[threadIdx.x^2 ];
      sywr[threadIdx.x] += sywr[threadIdx.x^1 ];
      sywi[threadIdx.x] += sywi[threadIdx.x^1 ];
    }
    if (threadIdx.x==0) {
       gri[blockIdx.x].x = sywr[0];
       gri[blockIdx.x].y = sywi[0];
    }
  }
}

static __global__ void 
abs2_kernel(const cuhqfield *y, float *g, const int *ieo){
//
// g = |y|^2  (ieo)
//
  float ywr = 0.0f;
#pragma unroll 6
  for(unsigned int i=0;i<CSH;i++) {
    __align__(16) float4 yri;
    yri = (*y).b[blockIdx.x].yri[i].c[threadIdx.x];
    ywr += yri.x * yri.x;
    ywr += yri.y * yri.y;
    ywr += yri.z * yri.z;
    ywr += yri.w * yri.w;
  }
#ifdef _SF
  {
    int itb    = thd_t + blk_t*_NBT;
    int iz     = thd_z + blk_z*_NBZ;
    int iy     = thd_y + blk_y*_NBY;
    int ix     = thd_x + blk_x*_NBX;
    int ieoxyz = ((*ieo)+(iz+1)+(iy+1)+(ix+1))%2;
    float flag = (float)(1-ieoxyz*((_NTH - itb)/_NTH));
    ywr = ywr * flag;
  }
#endif

  {
    __shared__ __align__(16) float a[CEIL];
#if THREADS != CEIL
    a[CEIL - THREADS + threadIdx.x] = 0.0f;
#endif

    __syncthreads();

    a[threadIdx.x] = ywr;

    __syncthreads();

    for (unsigned int s = CEIL/2; s > 32 ; s >>= 1){
      if (threadIdx.x < s) a[threadIdx.x] += a[threadIdx.x^s];
      __syncthreads();
    }
    if (threadIdx.x < 32) {
      a[threadIdx.x] += a[threadIdx.x^32];
      a[threadIdx.x] += a[threadIdx.x^16];
      a[threadIdx.x] += a[threadIdx.x^8 ];
      a[threadIdx.x] += a[threadIdx.x^4 ];
      a[threadIdx.x] += a[threadIdx.x^2 ];
      a[threadIdx.x] += a[threadIdx.x^1 ];
    }
    if (threadIdx.x==0) g[blockIdx.x] = a[0];
  }
}

static __global__ void 
reduce_float_kernel(float *g){
  unsigned int ith = threadIdx.x;
  unsigned int i   = blockIdx.x*(blockDim.x*2) + threadIdx.x;
  extern __shared__ __align__(16) float a[];

  a[ith] = g[i] + g[i+blockDim.x];

  __syncthreads();

  for (unsigned int s = blockDim.x/2; s > 0; s >>= 1){
    if (ith < s) a[ith] += a[ith^s];
    __syncthreads();
  }
  if (ith==0) g[blockIdx.x] = a[0];
}

static __global__ void 
reduce_complex_kernel(cuComplex *gri){
  unsigned int ith = threadIdx.x;
  unsigned int i   = blockIdx.x*(blockDim.x*2) + threadIdx.x;
  extern __shared__ __align__(16) cuComplex ari[];

  ari[ith].x = gri[i].x + gri[i+blockDim.x].x;
  ari[ith].y = gri[i].y + gri[i+blockDim.x].y;

  __syncthreads();

  for (unsigned int s = blockDim.x/2; s > 0; s >>= 1){
    if (ith < s) {
      ari[ith].x += ari[ith^s].x;
      ari[ith].y += ari[ith^s].y;
    }
    __syncthreads();
  }
  if (ith==0) gri[blockIdx.x] = ari[0];
}

static void 
cureverse_eo(const int *ieo,int *ioe){
  cureverse_eo_kernel<<<1,1>>>(ieo,ioe);
  CUT_CHECK_ERROR("cureverse_eo Kernel execution failed");
}

static float
abs2(const cuhqfield *y, const int *ieo){
//
// abs2 = |y|^2 (ieo)
//
  float a;
  float *gdat;
  CUDA_SAFE_CALL(cudaMalloc((void**)&gdat,sizeof(float)*BLOCKS));

  //
  // spin, color, TZ-thread reduction
  //
  abs2_kernel<<<BLOCKS,THREADS>>>(y,gdat,ieo);
  CUT_CHECK_ERROR("abs2 Kernel execution failed");

  //
  // remaining site reduction
  //
#if THREADS == CEIL
  {
    //
    // GPU side reduction
    //
    unsigned int ndat = BLOCKS;
    unsigned int threads = (ndat < maxThreads*2) ? nextPow2((ndat + 1)/ 2) : maxThreads;
    unsigned int blocks  = (ndat + threads*2 - 1) / (threads*2);
    dim3 dimGrid(blocks,1,1);
    dim3 dimBlock(threads,1,1);
    size_t size = threads*sizeof(float);

    reduce_float_kernel<<<dimGrid,dimBlock,size>>>(gdat);
    CUT_CHECK_ERROR("reduce_float Kernel execution failed");

#if CUDA_VERSION >= 4000
    CUDA_SAFE_CALL(cudaDeviceSynchronize());
#else
    CUDA_SAFE_CALL(cudaThreadSynchronize());
#endif

    float *hhdat = (float *)malloc(blocks*sizeof(float));
    CUDA_SAFE_CALL(cudaMemcpy(hhdat,gdat,blocks*sizeof(float),cudaMemcpyDeviceToHost));
    a = 0.0f;
    for (unsigned int i=0;i < blocks; i++) {
      a += hhdat[i];
    }
    free(hhdat);
  }
#else
  {
    //
    // host side reduction
    //

    float hdat[BLOCKS];
    CUDA_SAFE_CALL(cudaMemcpy(hdat,gdat,sizeof(float)*BLOCKS,cudaMemcpyDeviceToHost));

    a = 0.0f;
    for (unsigned int i=0;i < BLOCKS; i++) {
      a += hdat[i];
    }
  }
#endif

  CUDA_SAFE_CALL(cudaFree(gdat));
  return a;
}

static cuComplex
prod(const cuhqfield *y, const cuhqfield *w, const int *ieo){
//
// prod = <y|w>  (ieo)
//
  cuComplex p;
  cuComplex *gri;
  CUDA_SAFE_CALL(cudaMalloc((void**)&gri,sizeof(cuComplex)*BLOCKS));

  //
  // spin, color, TZ-thread reduction
  //
  prod_kernel<<<BLOCKS,THREADS>>>(y,w,gri,ieo);
  CUT_CHECK_ERROR("prod Kernel execution failed");

  //
  // remaining site reduction
  //
#if THREADS == CEIL
  {
    //
    // GPU side reduction
    //
    unsigned int ndat = BLOCKS;
    unsigned int threads = (ndat < maxThreads*2) ? nextPow2((ndat + 1)/ 2) : maxThreads;
    unsigned int blocks  = (ndat + threads*2 - 1) / (threads*2);
    dim3 dimGrid(blocks,1,1);
    dim3 dimBlock(threads,1,1);
    size_t size = threads*sizeof(cuComplex);

    reduce_complex_kernel<<<dimGrid,dimBlock,size>>>(gri);
    CUT_CHECK_ERROR("reduce_complex Kernel execution failed");

#if CUDA_VERSION >= 4000
    CUDA_SAFE_CALL(cudaDeviceSynchronize());
#else
    CUDA_SAFE_CALL(cudaThreadSynchronize());
#endif

    cuComplex *hhri = (cuComplex *)malloc(blocks*sizeof(cuComplex));
    CUDA_SAFE_CALL(cudaMemcpy(hhri,gri,blocks*sizeof(cuComplex),cudaMemcpyDeviceToHost));
    p.x = 0.0f;
    p.y = 0.0f;
    for (unsigned int i=0;i < blocks; i++) {
      p.x += hhri[i].x;
      p.y += hhri[i].y;
    }
    free(hhri);
  }
#else
  {
    //
    // host side reducion
    //
    cuComplex hri[BLOCKS];
    CUDA_SAFE_CALL(cudaMemcpy(hri,gri,sizeof(cuComplex)*BLOCKS,cudaMemcpyDeviceToHost));

    p.x = 0.0f;
    p.y = 0.0f;
    for (unsigned int i=0;i < BLOCKS; i++) {
      p.x += hri[i].x;
      p.y += hri[i].y;
    }
  }
#endif

  CUDA_SAFE_CALL(cudaFree(gri));
  return p;
}

static void 
assign(cuhqfield *a, const cuhqfield *b){
//
//  a = b
//
  CUDA_SAFE_CALL(cudaMemcpy( a, b, sizeof(cuhqfield), cudaMemcpyDeviceToDevice));
}

static void 
accum_sub(cuhqfield *a, const cuhqfield *b, const int *ieo){
//
// a = a - b
//
  accum_sub_kernel<<<BLOCKS,THREADS>>>(a,b);
  CUT_CHECK_ERROR("accum_sub Kernel execution failed");
}

static void 
accum_add_mult(cuhqfield *w, const cuComplex *ccoef, const cuhqfield *y){
//
// w = w + ccoef * y
//
  accum_add_mult_kernel<<<BLOCKS,THREADS>>>(w,ccoef,y);
  CUT_CHECK_ERROR("accum_add_multc Kernel execution failed");
}

static void 
accum_add_mult(cuhqfield *w, const float *rcoef, const cuhqfield *y){
//
// w = w + rcoef * y
//
  accum_add_mult_kernel<<<BLOCKS,THREADS>>>(w,rcoef,y);
  CUT_CHECK_ERROR("accum_add_multr Kernel execution failed");
}

static void 
accum_sub_mult(cuhqfield *w, const cuComplex *ccoef, const cuhqfield *y){
//
// w = w - ccoef * y
//
  accum_sub_mult_kernel<<<BLOCKS,THREADS>>>(w,ccoef,y);
  CUT_CHECK_ERROR("accum_sub_multc Kernel execution failed");
}

static void 
accum_sub_mult(cuhqfield *w, const float *rcoef, const cuhqfield *y){
//
// w = w - rcoef * y
//
  accum_sub_mult_kernel<<<BLOCKS,THREADS>>>(w,rcoef,y);
  CUT_CHECK_ERROR("accum_sub_multr Kernel execution failed");
}

static void 
accum_mult(cuhqfield *y, const float *rcoef){
//
// y = rcoef * y
//
  accum_mult_kernel<<<BLOCKS,THREADS>>>(y,rcoef);
  CUT_CHECK_ERROR("accum_mult Kernel execution failed");
}

static void 
accum_mult_add(cuhqfield *w, const float *rcoef, const cuhqfield *y){
//
// w = rcoef * w + y
//
  accum_mult_add_kernel<<<BLOCKS,THREADS>>>(w,rcoef,y);
  CUT_CHECK_ERROR("accum_mult_add Kernel execution failed");
}

static void 
accum_nmult_add(cuhqfield *w, const float *rcoef, const cuhqfield *y){
//
// w = -rcoef * w + y
//
  accum_nmult_add_kernel<<<BLOCKS,THREADS>>>(w,rcoef,y);
  CUT_CHECK_ERROR("accum_nmult_add Kernel execution failed");
}

static void 
accum_mult_add(cuhqfield *w, const cuComplex *ccoef, const cuhqfield *y){
//
// w = ccoef * w + y
//
  accum_mult_add_kernel<<<BLOCKS,THREADS>>>(w,ccoef,y);
  CUT_CHECK_ERROR("accum_mult_add Kernel execution failed");
}

static void 
accum_nmult_add(cuhqfield *w, const cuComplex *ccoef, const cuhqfield *y){
//
// w = -ccoef * w + y
//
  accum_nmult_add_kernel<<<BLOCKS,THREADS>>>(w,ccoef,y);
  CUT_CHECK_ERROR("accum_nmult_add Kernel execution failed");
}

static void
clear(cuhqfield *y){
  clear_kernel<<<BLOCKS,THREADS>>>(y);
  CUT_CHECK_ERROR("clear Kernel execution failed");
}

#undef  thd_t
#undef  thd_z
#undef  thd_y
#undef  thd_x
#undef  blk_t
#undef  blk_z
#undef  blk_y
#undef  blk_x

#endif
