#undef _CHECKSUM_

#define STRIDE_TZ                        (sizeof(cu_wqf_eo_thd)*_NBZ)
#define NGHOST_X(_depth_)                (sizeof(cu_wqf_eo_thd)*_NBZ*_NBY*(_depth_))
#define NGHOST_Y(_depth_)                (sizeof(cu_wqf_eo_thd)*_NBZ*(_NBX+(_depth_)*2)*_NGX*(_depth_))
#define BUFF_X_IDX(iby,ibx,igx,_depth_)  (((iby-1)+(ibx+igx*(_depth_))*_NBY)*STRIDE_TZ)
#define BUFF_Y_IDX(iby,ibx,igx,_depth_)  ((iby+((ibx-1+(_depth_))+igx*(_NBX+2*(_depth_)))*(_depth_))*STRIDE_TZ)

static int gpu_is_initialized = 0;

typedef struct _ALIG_ {
  char *send_buf;
  char *recv_buf;
  int   size;
  int   dst_rank;
  int   src_rank;
  MPI_Request sreq, rreq;
} gpu_combuf_t;

static gpu_combuf_t up_x,dn_x,up_y,dn_y;

static cudaStream_t stream_up_x_buff[_NGX][_NDEPTH][_NBY];
static cudaStream_t stream_dn_x_buff[_NGX][_NDEPTH][_NBY];
static cudaStream_t stream_up_y_buff[_NGX][_NDEPTH][_NBX+2*_NDEPTH];
static cudaStream_t stream_dn_y_buff[_NGX][_NDEPTH][_NBX+2*_NDEPTH];
//
// Buffer for ghost sites exchange. (allocated by cudaHostAlloc on pinned memory)
//
// GPU<=>CPU BUFF
static char *gbuff_x_up;
static char *gbuff_x_dn;
static char *gbuff_y_up;
static char *gbuff_y_dn;

// MPI BUFF
#if _NDIMX != 1
static char *sbuff_x_up;
static char *sbuff_x_dn;
static char *rbuff_x_up;
static char *rbuff_x_dn;
#endif
#if _NDIMY != 1
static char *sbuff_y_up;
static char *sbuff_y_dn;
static char *rbuff_y_up;
static char *rbuff_y_dn;
#endif

static long long int __call_counter = 0;

static inline long long int
__get_checksum(const int length, char *vec) {
  long long int csum;
  csum = 0;
  for (int i = 0; i < length; ++i) {
    csum = csum + (long long int)(vec[i]);
  }
  return csum;
}

static inline void
__sendrecv_char(const int size, char *sbuff, const int dst_rank,
                                char *rbuff, const int src_rank) {
#ifdef _QMX_
  qmx_sendrecv_char(size,sbuff,dst_rank,rbuff,src_rank);
#else
  MPI_Status sstat;
  MPI_Sendrecv((void *)sbuff,size,MPI_BYTE,dst_rank,_nodeid_     ,
               (void *)rbuff,size,MPI_BYTE,src_rank,src_rank,MPI_COMM_WORLD,&sstat);
#endif
#if defined(_CHECKSUM_)
  long long int ssum = __get_checksum(size,sbuff);
  long long int rsum = __get_checksum(size,rbuff);
  printf("%10lld: size:%12d send:%2d -> %2d %12lld  recv:%d -> %d %12lld\n",
     __call_counter,size,_nodeid_,dst_rank,ssum,src_rank,_nodeid_,rsum);
  __call_counter++;
#endif
}

static inline void
gpu_combuf_isendrecv_char(gpu_combuf_t *combuf, const int size,
                          char *sbuff, const int dst_rank,
                          char *rbuff, const int src_rank) {
  combuf->send_buf = sbuff;
  combuf->recv_buf = rbuff;
  combuf->size     = size;
  combuf->dst_rank = dst_rank;
  combuf->src_rank = src_rank;
  MPI_Irecv((void *)(combuf->recv_buf),size,MPI_BYTE,src_rank,src_rank,MPI_COMM_WORLD,&(combuf->rreq));
  MPI_Isend((void *)(combuf->send_buf),size,MPI_BYTE,dst_rank,_nodeid_,MPI_COMM_WORLD,&(combuf->sreq));
}

static inline void
gpu_combuf_check_char(gpu_combuf_t *combuf) {
  MPI_Status  stat[2];
  MPI_Request  req[2];
  req[0] = combuf->rreq;
  req[1] = combuf->sreq;
  MPI_Waitall(2,req,stat);
}

static inline void
gpu_combuf_check_char(gpu_combuf_t *combuf_A, gpu_combuf_t *combuf_B) {
  MPI_Status  stat[4];
  MPI_Request  req[4];
  req[0] = combuf_A->rreq;
  req[1] = combuf_A->sreq;
  req[2] = combuf_B->rreq;
  req[3] = combuf_B->sreq;
  MPI_Waitall(4,req,stat);
}

static void
new_copy_boundary() {
//
// create ghost site buffers
//

  if (0 == gpu_is_initialized) {

#if CUDA_VERSION >= 4000


#pragma omp parallel for
    for (int igx = 0; igx < _NGX; igx++) {
      switch_gpu_(&igx);
      for (int ibx = 0; ibx < _NDEPTH; ibx++) {
      for (int iby = 0; iby < _NBY; iby++) {
        CUDA_SAFE_CALL(cudaStreamCreate(&(stream_up_x_buff[igx][ibx][iby])));
        CUDA_SAFE_CALL(cudaStreamCreate(&(stream_dn_x_buff[igx][ibx][iby])));
      }
      }
      for (int iby = 0; iby < _NDEPTH; iby++) {
      for (int ibx = 0; ibx < _NBX + 2*_NDEPTH; ibx++) {
        CUDA_SAFE_CALL(cudaStreamCreate(&(stream_up_y_buff[igx][iby][ibx])));
        CUDA_SAFE_CALL(cudaStreamCreate(&(stream_dn_y_buff[igx][iby][ibx])));
      }
      }
    }
#endif

    //
    // allocates memory for 1st call.
    //

    //
    // Pinned memory on host side
    //
#if CUDA_VERSION >= 4000
    CUDA_SAFE_CALL(cudaHostAlloc((void **)&gbuff_x_up,NGHOST_X(_NDEPTH)*(_NGX+1),cudaHostAllocPortable));
    CUDA_SAFE_CALL(cudaHostAlloc((void **)&gbuff_x_dn,NGHOST_X(_NDEPTH)*(_NGX+1),cudaHostAllocPortable));
    CUDA_SAFE_CALL(cudaHostAlloc((void **)&gbuff_y_up,NGHOST_Y(_NDEPTH)*2   ,cudaHostAllocPortable));
    CUDA_SAFE_CALL(cudaHostAlloc((void **)&gbuff_y_dn,NGHOST_Y(_NDEPTH)*2   ,cudaHostAllocPortable));
#else
    CUDA_SAFE_CALL(cudaHostAlloc((void **)&gbuff_x_up,NGHOST_X(_NDEPTH)*_NGX,cudaHostAllocPortable));
    CUDA_SAFE_CALL(cudaHostAlloc((void **)&gbuff_x_dn,NGHOST_X(_NDEPTH)*_NGX,cudaHostAllocPortable));
    CUDA_SAFE_CALL(cudaHostAlloc((void **)&gbuff_y_up,NGHOST_Y(_NDEPTH)     ,cudaHostAllocPortable));
    CUDA_SAFE_CALL(cudaHostAlloc((void **)&gbuff_y_dn,NGHOST_Y(_NDEPTH)     ,cudaHostAllocPortable));
#endif

#if CUDA_VERSION >= 4000
    memset((void *)gbuff_x_up,0,NGHOST_X(_NDEPTH)*(_NGX+1));
    memset((void *)gbuff_x_dn,0,NGHOST_X(_NDEPTH)*(_NGX+1));
    memset((void *)gbuff_y_up,0,NGHOST_Y(_NDEPTH)*2);
    memset((void *)gbuff_y_dn,0,NGHOST_Y(_NDEPTH)*2);
#else
    memset(gbuff_x_up,0,NGHOST_X(_NDEPTH)*_NGX);
    memset(gbuff_x_dn,0,NGHOST_X(_NDEPTH)*_NGX);
    memset(gbuff_y_up,0,NGHOST_Y(_NDEPTH));
    memset(gbuff_y_dn,0,NGHOST_Y(_NDEPTH));
#endif

#if CUDA_VERSION >= 4000
    // NO MPI buffers
#else
    //
    // MPI buffers
    //
#if _NDIMX != 1
    sbuff_x_up = (char *)_mm_malloc(NGHOST_X(_NDEPTH),16);
    sbuff_x_dn = (char *)_mm_malloc(NGHOST_X(_NDEPTH),16);
    rbuff_x_up = (char *)_mm_malloc(NGHOST_X(_NDEPTH),16);
    rbuff_x_dn = (char *)_mm_malloc(NGHOST_X(_NDEPTH),16);
    memset(sbuff_x_up,0,NGHOST_X(_NDEPTH));
    memset(sbuff_x_dn,0,NGHOST_X(_NDEPTH));
    memset(rbuff_x_up,0,NGHOST_X(_NDEPTH));
    memset(rbuff_x_dn,0,NGHOST_X(_NDEPTH));
#endif
#if _NDIMY != 1
    sbuff_y_up = (char *)_mm_malloc(NGHOST_Y(_NDEPTH),16);
    sbuff_y_dn = (char *)_mm_malloc(NGHOST_Y(_NDEPTH),16);
    rbuff_y_up = (char *)_mm_malloc(NGHOST_Y(_NDEPTH),16);
    rbuff_y_dn = (char *)_mm_malloc(NGHOST_Y(_NDEPTH),16);
    memset(sbuff_y_up,0,NGHOST_X(_NDEPTH));
    memset(sbuff_y_dn,0,NGHOST_X(_NDEPTH));
    memset(rbuff_y_up,0,NGHOST_X(_NDEPTH));
    memset(rbuff_y_dn,0,NGHOST_X(_NDEPTH));
#endif
#endif

    gpu_is_initialized = 1;
  }
}

static void
delete_copy_boundary() {
//
// delete ghost site buffers
//

/*
  CUDA_SAFE_CALL(cudaFreeHost(gbuff_x_up));
  CUDA_SAFE_CALL(cudaFreeHost(gbuff_x_dn));
  CUDA_SAFE_CALL(cudaFreeHost(gbuff_y_up));
  CUDA_SAFE_CALL(cudaFreeHost(gbuff_y_dn));

#if _NDIMX != 1
  _mm_free(sbuff_x_up);
  _mm_free(sbuff_x_dn);
  _mm_free(rbuff_x_up);
  _mm_free(rbuff_x_dn);
#endif
#if _NDIMY != 1
  _mm_free(sbuff_y_up);
  _mm_free(sbuff_y_dn);
  _mm_free(rbuff_y_up);
  _mm_free(rbuff_y_dn);
#endif
*/

}

static void 
copy_boundary_send_x(const int depth, cu_wqf_eo *ye){
//
// copy boundary in x direction (GPU=>HOST=>GPU send phase in a node)
//
#if CUDA_VERSION >= 4000
  // GPU => HOST => GPU (step 1)

#if _NDIMX == 1

#pragma omp parallel for
  for (int igx = 0; igx < _NGX; igx++) {
    switch_gpu_(&igx);
    int igx_up = (igx+1+_NGX) % _NGX;
    int igx_dn = (igx-1+_NGX) % _NGX;
    for (int ibx = 0; ibx <= depth-1; ibx++) {
    for (int iby = 1; iby <= _NBY; iby++) {
      cudaMemcpyPeerAsync(&((*ye).g[igx_up]->blk[blk_idx(1,iby,     1-depth+ibx)]),igx_up,
                          &((*ye).g[igx   ]->blk[blk_idx(1,iby,_NBX-depth+1+ibx)]),igx,
                          STRIDE_TZ,stream_up_x_buff[igx][ibx][iby-1]);
      cudaMemcpyPeerAsync(&((*ye).g[igx_dn]->blk[blk_idx(1,iby,      _NBX+1+ibx)]),igx_dn,
                          &((*ye).g[igx   ]->blk[blk_idx(1,iby,           1+ibx)]),igx,
                          STRIDE_TZ,stream_dn_x_buff[igx][ibx][iby-1]);

    }
    }
    cudaDeviceSynchronize();
  }

#else

#pragma omp parallel for
  for (int igx = 0; igx < _NGX; igx++) {
    switch_gpu_(&igx);
    int igx_up = (igx+1);
    for (int ibx = 0; ibx <= depth-1; ibx++) {
    for (int iby = 1; iby <= _NBY; iby++) {
      cudaMemcpyAsync(&(gbuff_x_up[BUFF_X_IDX(iby,ibx,igx_up,depth)]),
                      &((*ye).g[igx]->blk[blk_idx(1,iby,_NBX-depth+1+ibx)]),
                      STRIDE_TZ,cudaMemcpyDeviceToHost,stream_up_x_buff[igx][ibx][iby-1]);
      cudaMemcpyAsync(&(gbuff_x_dn[BUFF_X_IDX(iby,ibx,igx   ,depth)]),
                      &((*ye).g[igx]->blk[blk_idx(1,iby,           1+ibx)]),
                      STRIDE_TZ,cudaMemcpyDeviceToHost,stream_dn_x_buff[igx][ibx][iby-1]);
    }
    }
    cudaDeviceSynchronize();
  }
#endif

#else

  // GPU => CPU => GPU (step 1) (within a node)
#pragma omp parallel for
  for (int igx = 0; igx < _NGX; igx++) {
    switch_gpu_(&igx);
    for (int ibx = 0; ibx <= depth-1; ibx++) {
    for (int iby = 1; iby <= _NBY; iby++) {
      CUDA_SAFE_CALL(cudaMemcpy(&(gbuff_x_up[BUFF_X_IDX(iby,ibx,igx,depth)]),
                                &((*ye).g[igx]->blk[blk_idx(1,iby,_NBX-depth+1+ibx)]),
                                STRIDE_TZ,cudaMemcpyDeviceToHost));
      CUDA_SAFE_CALL(cudaMemcpy(&(gbuff_x_dn[BUFF_X_IDX(iby,ibx,igx,depth)]),
                                &((*ye).g[igx]->blk[blk_idx(1,iby,           1+ibx)]),
                                STRIDE_TZ,cudaMemcpyDeviceToHost));
    }
    }
  }
#pragma omp parallel for
  for (int igx = 0; igx < _NGX; igx++) {
    switch_gpu_(&igx);
    int igx_up = (igx+1+_NGX) % _NGX;
    int igx_dn = (igx-1+_NGX) % _NGX;
    for (int ibx = 0; ibx <= depth-1; ibx++) {
    for (int iby = 1; iby <= _NBY; iby++) {
      CUDA_SAFE_CALL(cudaMemcpy(&((*ye).g[igx]->blk[blk_idx(1,iby,     1-depth+ibx)]),
                                &(gbuff_x_up[BUFF_X_IDX(iby,ibx,igx_dn,depth)]),
                                STRIDE_TZ,cudaMemcpyHostToDevice));
      CUDA_SAFE_CALL(cudaMemcpy(&((*ye).g[igx]->blk[blk_idx(1,iby,      _NBX+1+ibx)]),
                                &(gbuff_x_dn[BUFF_X_IDX(iby,ibx,igx_up,depth)]),
                                STRIDE_TZ,cudaMemcpyHostToDevice));
    }
    }
  }

#endif 

}

static void
copy_boundary_sendrecv_x(const int depth){
//
// copy boundary in x direction (HOST<=>HOST MPI sendrecv phase)
//

#if _NDIMX != 1

#if CUDA_VERSION >= 4000

  gpu_combuf_isendrecv_char(&up_x, NGHOST_X(depth),
                            (char *)&(gbuff_x_up[NGHOST_X(depth)*(_NGX)]),_nodeidup_(0),
                            (char *)&(gbuff_x_up[                     0]),_nodeiddn_(0));
  gpu_combuf_isendrecv_char(&dn_x, NGHOST_X(depth),
                            (char *)&(gbuff_x_dn[                     0]),_nodeiddn_(0),
                            (char *)&(gbuff_x_dn[NGHOST_X(depth)*(_NGX)]),_nodeidup_(0));
/*
  __sendrecv_char(NGHOST_X(depth),(char *)&(gbuff_x_up[NGHOST_X(depth)*(_NGX)]),_nodeidup_(0),
                                  (char *)&(gbuff_x_up[                     0]),_nodeiddn_(0));
  __sendrecv_char(NGHOST_X(depth),(char *)&(gbuff_x_dn[                     0]),_nodeiddn_(0),
                                  (char *)&(gbuff_x_dn[NGHOST_X(depth)*(_NGX)]),_nodeidup_(0));
*/

#else

#pragma omp parallel sections
  {
#pragma omp section
    memcpy((void *)&(sbuff_x_up[0]),(void *)&(gbuff_x_up[NGHOST_X(depth)*(_NGX-1)]),NGHOST_X(depth));
#pragma omp section
    memcpy((void *)&(sbuff_x_dn[0]),(void *)&(gbuff_x_dn[                       0]),NGHOST_X(depth));
  }

  __sendrecv_char(NGHOST_X(depth),sbuff_x_up,_nodeidup_(0),rbuff_x_up,_nodeiddn_(0));
  __sendrecv_char(NGHOST_X(depth),sbuff_x_dn,_nodeiddn_(0),rbuff_x_dn,_nodeidup_(0));

#pragma omp parallel sections
  {
#pragma omp section
    memcpy((void *)&(gbuff_x_up[NGHOST_X(depth)*(_NGX-1)]),(void *)&(rbuff_x_up[0]),NGHOST_X(depth));
#pragma omp section
    memcpy((void *)&(gbuff_x_dn[                       0]),(void *)&(rbuff_x_dn[0]),NGHOST_X(depth));
  }

#endif

#endif
}

static void
copy_boundary_recv_x(const int depth, cu_wqf_eo *ye){
//
// copy boundary in x direction (HOST=>GPU recv phase)
//
#if CUDA_VERSION >= 4000

#if _NDIMX != 1
  gpu_combuf_check_char(&up_x,&dn_x);

#pragma omp parallel for
  for (int igx = 0; igx < _NGX; igx++) {
    switch_gpu_(&igx);
    int igx_up = (igx+1);
    for (int ibx = 0; ibx <= depth-1; ibx++) {
    for (int iby = 1; iby <= _NBY; iby++) {
      cudaMemcpyAsync(&((*ye).g[igx]->blk[blk_idx(1,iby,     1-depth+ibx)]),
                      &(gbuff_x_up[BUFF_X_IDX(iby,ibx,igx   ,depth)]),
                      STRIDE_TZ,cudaMemcpyHostToDevice,stream_up_x_buff[igx][ibx][iby-1]);
      cudaMemcpyAsync(&((*ye).g[igx]->blk[blk_idx(1,iby,      _NBX+1+ibx)]),
                      &(gbuff_x_dn[BUFF_X_IDX(iby,ibx,igx_up,depth)]),
                      STRIDE_TZ,cudaMemcpyHostToDevice,stream_dn_x_buff[igx][ibx][iby-1]);
    }
    }
    cudaDeviceSynchronize();
  }
#endif

#else

#if _NDIMX != 1
  // CPU => GPU (step 4)
#pragma omp parallel for
  for (int igx=0;igx<_NGX;igx++) {
    switch_gpu_(&igx);
    int igx_up = (igx+1+_NGX) % _NGX;
    int igx_dn = (igx-1+_NGX) % _NGX;
    if (0==igx) {
      for (int ibx = 0; ibx <= depth-1; ibx++) {
      for (int iby = 1; iby <= _NBY; iby++) {
        CUDA_SAFE_CALL(cudaMemcpy(&((*ye).g[igx]->blk[blk_idx(1,iby,     1-depth+ibx)]),
                                  &(gbuff_x_up[BUFF_X_IDX(iby,ibx,igx_dn,depth)]),
                                  STRIDE_TZ,cudaMemcpyHostToDevice));
      }
      }
    }
    if (_NGX-1==igx) {
      for (int ibx = 0; ibx <= depth-1; ibx++) {
      for (int iby = 1; iby <= _NBY; iby++) {
        CUDA_SAFE_CALL(cudaMemcpy(&((*ye).g[igx]->blk[blk_idx(1,iby,      _NBX+1+ibx)]),
                                  &(gbuff_x_dn[BUFF_X_IDX(iby,ibx,igx_up,depth)]),
                                  STRIDE_TZ,cudaMemcpyHostToDevice));
      }
      }
    }
  }
#endif

#endif
  ycopy_dsize += NGHOST_X(depth)*2;
}


static void
copy_boundary_send_y(const int depth, cu_wqf_eo *ye){
//
// copy boundary in y direction (GPU=>HOST=>GPU send phase in a node)
//

#if CUDA_VERSION >= 4000

#if _NDIMY == 1

#pragma omp parallel for
  for (int igx = 0; igx < _NGX;igx++) {
    switch_gpu_(&igx);
    for (int ibx = 1-depth; ibx <= _NBX+depth;ibx++) {
    for (int iby = 0; iby <= depth-1;iby++) {
      cudaMemcpyAsync(&((*ye).g[igx]->blk[blk_idx(1,     1-depth+iby,ibx)]),
                      &((*ye).g[igx]->blk[blk_idx(1,_NBY+1-depth+iby,ibx)]),
                      STRIDE_TZ,cudaMemcpyDeviceToDevice, stream_up_y_buff[igx][iby][ibx-1+depth]);
      cudaMemcpyAsync(&((*ye).g[igx]->blk[blk_idx(1,      _NBY+1+iby,ibx)]),
                      &((*ye).g[igx]->blk[blk_idx(1,           1+iby,ibx)]),
                      STRIDE_TZ,cudaMemcpyDeviceToDevice, stream_dn_y_buff[igx][iby][ibx-1+depth]);
    }
    }
    cudaDeviceSynchronize();
  }

#else

// GPU => GPU (to another HOST) (3 steps)
  // GPU => CPU (step 1)
#pragma omp parallel for
  for (int igx = 0;igx < _NGX;igx++) {
    switch_gpu_(&igx);
    for (int ibx = 1-depth; ibx <= _NBX+depth;ibx++) {
    for (int iby = 0; iby <= depth-1;iby++) {
      cudaMemcpyAsync(&(gbuff_y_up[BUFF_Y_IDX(iby,ibx,igx,depth)]),
                      &((*ye).g[igx]->blk[blk_idx(1,_NBY-depth+1+iby,ibx)]),
                      STRIDE_TZ,cudaMemcpyDeviceToHost, stream_up_y_buff[igx][iby][ibx-1+depth]);
      cudaMemcpyAsync(&(gbuff_y_dn[BUFF_Y_IDX(iby,ibx,igx,depth)]),
                      &((*ye).g[igx]->blk[blk_idx(1,           1+iby,ibx)]),
                      STRIDE_TZ,cudaMemcpyDeviceToHost, stream_dn_y_buff[igx][iby][ibx-1+depth]);
    }
    }
    cudaDeviceSynchronize();
  }
#endif


#else

#pragma omp parallel for
  for (int igx = 0; igx < _NGX;igx++) {
    switch_gpu_(&igx);
    for (int ibx = 1-depth; ibx <= _NBX+depth;ibx++) {
    for (int iby = 0; iby <= depth-1;iby++) {
      CUDA_SAFE_CALL(cudaMemcpy(&((*ye).g[igx]->blk[blk_idx(1,     1-depth+iby,ibx)]),
                                &((*ye).g[igx]->blk[blk_idx(1,_NBY+1-depth+iby,ibx)]),
                                STRIDE_TZ,cudaMemcpyDeviceToDevice));
      CUDA_SAFE_CALL(cudaMemcpy(&((*ye).g[igx]->blk[blk_idx(1,      _NBY+1+iby,ibx)]),
                                &((*ye).g[igx]->blk[blk_idx(1,           1+iby,ibx)]),
                                STRIDE_TZ,cudaMemcpyDeviceToDevice));
    }
    }
  }


#if _NDIMY != 1
// GPU => GPU (to another HOST) (3 steps)
  // GPU => CPU (step 1)
#pragma omp parallel for
  for (int igx = 0;igx < _NGX;igx++) {
    switch_gpu_(&igx);
    for (int ibx = 1-depth; ibx <= _NBX+depth;ibx++) {
    for (int iby = 0; iby <= depth-1;iby++) {
      CUDA_SAFE_CALL(cudaMemcpy(&(gbuff_y_up[BUFF_Y_IDX(iby,ibx,igx,depth)]),
                                &((*ye).g[igx]->blk[blk_idx(1,_NBY-depth+1+iby,ibx)]),
                                STRIDE_TZ,cudaMemcpyDeviceToHost));
      CUDA_SAFE_CALL(cudaMemcpy(&(gbuff_y_dn[BUFF_Y_IDX(iby,ibx,igx,depth)]),
                                &((*ye).g[igx]->blk[blk_idx(1,           1+iby,ibx)]),
                                STRIDE_TZ,cudaMemcpyDeviceToHost));
    }
    }
  }
#endif


#endif

}

static void
copy_boundary_sendrecv_y(const int depth){
//
// copy boundary in y direction (HOST<=>HOST MPI sendrecv phase)
//
//

#if _NDIMY != 1

#if CUDA_VERSION >= 4000

  gpu_combuf_isendrecv_char(&up_y,NGHOST_Y(depth),
                            gbuff_y_up                ,_nodeidup_(1),
                            gbuff_y_up+NGHOST_Y(depth),_nodeiddn_(1));
  gpu_combuf_isendrecv_char(&dn_y,NGHOST_Y(depth),
                            gbuff_y_dn                ,_nodeiddn_(1),
                            gbuff_y_dn+NGHOST_Y(depth),_nodeidup_(1));

/*
  __sendrecv_char(NGHOST_Y(depth),gbuff_y_up                ,_nodeidup_(1),
                                  gbuff_y_up+NGHOST_Y(depth),_nodeiddn_(1));
  __sendrecv_char(NGHOST_Y(depth),gbuff_y_dn                ,_nodeiddn_(1),
                                  gbuff_y_dn+NGHOST_Y(depth),_nodeidup_(1));
*/

#else

#pragma omp parallel sections
  {
#pragma omp section
    memcpy((void *)sbuff_y_up,(void *)gbuff_y_up,NGHOST_Y(depth));
#pragma omp section
    memcpy((void *)sbuff_y_dn,(void *)gbuff_y_dn,NGHOST_Y(depth));
  }

  __sendrecv_char(NGHOST_Y(depth),sbuff_y_up,_nodeidup_(1),rbuff_y_up,_nodeiddn_(1));
  __sendrecv_char(NGHOST_Y(depth),sbuff_y_dn,_nodeiddn_(1),rbuff_y_dn,_nodeidup_(1));

#pragma omp parallel sections
  {
#pragma omp section
    memcpy((void *)gbuff_y_up,(void *)rbuff_y_up,NGHOST_Y(depth));
#pragma omp section
    memcpy((void *)gbuff_y_dn,(void *)rbuff_y_dn,NGHOST_Y(depth));
  }
#endif

#endif

}

static void
copy_boundary_recv_y(const int depth, cu_wqf_eo *ye){
//
// copy boundary in x direction (HOST=>GPU recv phase)
//

#if _NDIMY != 1

  // CPU => GPU (step 3)

#if CUDA_VERSION >= 4000

  gpu_combuf_check_char(&up_y,&dn_y);

#pragma omp parallel for
  for (int igx = 0;igx < _NGX;igx++) {
    switch_gpu_(&igx);
    for (int ibx = 1-depth; ibx <= _NBX+depth;ibx++) {
    for (int iby = 0; iby <= depth-1;iby++) {
      cudaMemcpyAsync(&((*ye).g[igx]->blk[blk_idx(1,     1-depth+iby,ibx)]),
                      &(gbuff_y_up[BUFF_Y_IDX(iby,ibx,igx,depth)+NGHOST_Y(depth)]),
                      STRIDE_TZ,cudaMemcpyHostToDevice,stream_up_y_buff[igx][iby][ibx-1+depth]);
      cudaMemcpyAsync(&((*ye).g[igx]->blk[blk_idx(1,      _NBY+1+iby,ibx)]),
                      &(gbuff_y_dn[BUFF_Y_IDX(iby,ibx,igx,depth)+NGHOST_Y(depth)]),
                      STRIDE_TZ,cudaMemcpyHostToDevice,stream_dn_y_buff[igx][iby][ibx-1+depth]);
    }
    }
    cudaDeviceSynchronize();
  }

#else

#pragma omp parallel for
  for (int igx = 0;igx < _NGX;igx++) {
    switch_gpu_(&igx);
    for (int ibx = 1-depth; ibx <= _NBX+depth;ibx++) {
    for (int iby = 0; iby <= depth-1;iby++) {
      CUDA_SAFE_CALL(cudaMemcpy(&((*ye).g[igx]->blk[blk_idx(1,     1-depth+iby,ibx)]),
                                &(gbuff_y_up[BUFF_Y_IDX(iby,ibx,igx,depth)]),
                                STRIDE_TZ,cudaMemcpyHostToDevice));
      CUDA_SAFE_CALL(cudaMemcpy(&((*ye).g[igx]->blk[blk_idx(1,      _NBY+1+iby,ibx)]),
                                &(gbuff_y_dn[BUFF_Y_IDX(iby,ibx,igx,depth)]),
                                STRIDE_TZ,cudaMemcpyHostToDevice));
    }
    }
  }

#endif

#endif
  ycopy_dsize += NGHOST_Y(depth)*2;
}

static void
copy_boundary_all(cu_wqf_eo *ye){
//
// copy boundary incuding extended sites
//

  startTimer(ycopy_timer);

#ifndef _singlePU
#ifdef _QMX_
  qmx_barrier();
#else
  MPI_Barrier(MPI_COMM_WORLD);
#endif
#endif

  copy_boundary_send_x(_NDEPTH,ye);
  copy_boundary_sendrecv_x(_NDEPTH);
  copy_boundary_recv_x(_NDEPTH,ye);

  copy_boundary_send_y(_NDEPTH,ye);
  copy_boundary_sendrecv_y(_NDEPTH);
  copy_boundary_recv_y(_NDEPTH,ye);

  ycopy_count += 1;
  stopTimer(ycopy_timer);
}

static void
copy_boundary(cu_wqf_eo *ye){
//
// copy boundary for boundary edge
//

  startTimer(ycopy_timer);

#ifndef _singlePU
#ifdef _QMX_
  qmx_barrier();
#else
  MPI_Barrier(MPI_COMM_WORLD);
#endif
#endif

  copy_boundary_send_x(1,ye);
  copy_boundary_sendrecv_x(1);
  copy_boundary_recv_x(1,ye);

  copy_boundary_send_y(1,ye);
  copy_boundary_sendrecv_y(1);
  copy_boundary_recv_y(1,ye);

  ycopy_count += 1;
  stopTimer(ycopy_timer);
}

#undef STRIDE_TZ
#undef NGHOST_X
#undef NGHOST_Y
#undef BUFF_X_IDX
#undef BUFF_Y_IDX
