#!/usr/bin/ruby
#
# $Id: FieldClasses.rb,v 1.2 2010/09/01 04:20:07 ishikawa Exp $
#
require 'complex'

class Spinor

  J = Complex(0,1)

  @@col  = 3
  @@spin = 4
  @@eps  = 0.000001
  @@igamma = [[1,2,3,4],
              [4,3,2,1],
              [4,3,2,1],
              [3,4,1,2],
              [1,2,3,4],
              [3,4,1,2]]
  @@zgamma = [[ 1, 1, 1, 1],
              [-J,-J,+J,+J],
              [-1, 1, 1,-1],
              [-J,+J,+J,-J],
              [ 1, 1,-1,-1],
              [ 1, 1, 1, 1]]
  @@mudir = ["","x","y","z","t","w"]

  def initialize(myname)
    @y = Array.new(4) {|i| Complex(rand, rand)}
    @myname = myname
  end

  def set_name(myname)
    @y = Array.new(4) {|i| Complex(rand, rand)}
    @myname = myname
  end

  def get_name
    return @myname
  end

  def assign(y)
    return "#{@myname}(:,:) = #{y}(:,:)"
  end

  def assign_add(x,y)
    str = Array.new(4,nil)
    for i in 1..@@spin
      str[i-1] = "#{@myname}(:,#{i}) = #{x.get_name}(:,#{i}) + #{y.get_name}(:,#{i})"
    end
    return str
  end

  def assign_add_mult(x,y,coef)
    str = Array.new(4,nil)
    for i in 1..@@spin
      str[i-1] = "#{@myname}(:,#{i}) = #{x.get_name}(:,#{i}) + #{y.get_name}(:,#{i}) * #{coef}"
    end
    return str
  end

  def accum_add(y)
    str = Array.new(4,nil)
    for i in 1..@@spin
      str[i-1] = "#{@myname}(:,#{i}) = #{@myname}(:,#{i}) + #{y.get_name}(:,#{i})"
    end
    return str
  end

  def accum_add_mult(y,coef)
    str = Array.new(4,nil)
    for i in 1..@@spin
      str[i-1] = "#{@myname}(:,#{i}) = #{@myname}(:,#{i}) + #{y.get_name}(:,#{i}) * #{coef}"
    end
    return str
  end

  def get_mult_gamma(mu,y)
    str = Array.new(4,nil)
    for i in 1..@@spin
      if (mu > 0)
        sign = +1
      else
        sign = -1
      end
      jcoef = sign * @@zgamma[mu.abs][i-1]
      idx   =        @@igamma[mu.abs][i-1]
      if (jcoef == +1)
        str[i-1] = "+ #{y.get_name}(:,#{idx})"
      elsif (jcoef == -1)
        str[i-1] = "- #{y.get_name}(:,#{idx})"
      elsif (jcoef == +J)
        str[i-1] = "+ cmplx(-real(#{y.get_name}(:,#{idx})), &\n" +
                   "  &                aimag(#{y.get_name}(:,#{idx})),kind=DP)"
      elsif (jcoef == -J)
        str[i-1] = "- cmplx(-real(#{y.get_name}(:,#{idx})), &\n" +
                   "  &                aimag(#{y.get_name}(:,#{idx})),kind=DP)"
      end
    end
    return str
  end
  private :get_mult_gamma

  def assign_mult_gamma(mu,y)
    if (!@myname)
      puts "set my name!"
      abort
    end
    str = get_mult_gamma(mu,y)
    out_str = Array.new(4,nil)
    for i in 1..@@spin
      out_str[i-1] = "#{@myname}(:,#{i}) = " + str[i-1]
    end
    return out_str.compact
  end

  def accum_mult_gamma(mu,y)
    if (!@myname)
      puts "set my name!"
      abort
    end
    str = get_mult_gamma(mu,y)
    out_str = Array.new(4,nil)
    for i in 1..@@spin
      out_str[i-1] = "#{@myname}(:,#{i}) = #{@myname}(:,#{i}) " + str[i-1]
    end
    return out_str.compact
  end

  def assign_proj_up(mu,y)
    if (!@myname)
      puts "set my name!"
      abort
    end
    str,py = get_proj_up(mu,y)
    out_str = Array.new(4,nil)
    for i in 1..(@@spin/2)
      if (str[i-1])
        out_str[i-1] = "#{@myname}(:,#{i}) = " + str[i-1]
      else
        out_str[i-1] = "#{@myname}(:,#{i}) = " + str[i-1+@@spin/2]
      end
    end
    return out_str.compact
  end

  def get_proj_up(mu,y)
    if (mu > 0) 
      sign = +1
    elsif (mu < 0) 
      sign = -1
    end
    py = Array.new(4,Complex(0,0))
    str = Array.new(4,nil)

    for i in 1..@@spin
      jcoef = sign*@@zgamma[mu.abs][i-1]
      idx   =      @@igamma[mu.abs][i-1]
      py[i-1] = @y[i-1] + jcoef * @y[idx-1]

      if (idx != i)
        if (jcoef == 1)
          str[i-1] = "#{y.get_name}(:,#{i}) &\n" +
                     "  &        + #{y.get_name}(:,#{idx})"
        end
        if (jcoef == -1)
          str[i-1] = "#{y.get_name}(:,#{i}) &\n" +
                     "  &        - #{y.get_name}(:,#{idx})"
        end
        if (jcoef == J)
          str[i-1] = "#{y.get_name}(:,#{i}) &\n" +
                     "  &  + cmplx(-aimag(#{y.get_name}(:,#{idx})),  &\n" +
                     "  &            real(#{y.get_name}(:,#{idx})),kind=DP)"
        end
        if (jcoef == -J)
          str[i-1] = "#{y.get_name}(:,#{i}) &\n" + 
                     "  &  - cmplx(-aimag(#{y.get_name}(:,#{idx})),  &\n" +
                     "  &            real(#{y.get_name}(:,#{idx})),kind=DP)"
        end
      else
        if (jcoef ==  1)
          str[i-1] = "#{y.get_name}(:,#{i})*2.0_DP"
        end
        if (jcoef == -1)
          str[i-1] = nil
        end
      end

      if str[i-1]
        if (mu.abs == 5)
          str[i-1] = "("+str[i-1]+")*fac"
        end
      end

      if (py[i-1].abs < @@eps)
        py[i-1] = 0
      end

    end
    return str,py
  end
  private :get_proj_up

  def get_proj_low(mu,y)

    str,py = get_proj_up(mu,y)

    out_str = Array.new(4,nil)

    for i in 1..@@spin/2
      idx = @@igamma[mu.abs][i-1]

      if (str[i-1] && idx == i)
      else
        out_str[i-1         ] = "+ Z0"
        out_str[i-1+@@spin/2] = "+ #{y.get_name}(:,#{i})"
      end

      if (str[i-1] && idx != i)
        out_str[i-1         ] = "+ #{y.get_name}(:,#{i})"
      end
    end

    for i in (@@spin/2+1)..@@spin
      idx   = @@igamma[mu.abs][i-1]

      if (str[i-1] && idx == i)
      else
        out_str[i-1         ] = "+ Z0"
        out_str[i-1-@@spin/2] = "+ #{y.get_name}(:,#{i-@@spin/2})"
      end

      if (py[i-1].abs > @@eps && idx != i)
        ccoef = (py[i-1]/py[idx-1])
        cr = ccoef.real
        ci = ccoef.imag
        if (cr.abs > @@eps)
          if (cr > 0)
            out_str[i-1] = "+ #{y.get_name}(:,#{idx})"
          else
            out_str[i-1] = "- #{y.get_name}(:,#{idx})"
          end
        elsif (ci.abs > @@eps)
          if (ci > 0)
            out_str[i-1] = "+ cmplx(-aimag(#{y.get_name}(:,#{idx})),real(#{y.get_name}(:,#{idx})),kind=DP)"
          else
            out_str[i-1] = "- cmplx(-aimag(#{y.get_name}(:,#{idx})),real(#{y.get_name}(:,#{idx})),kind=DP)"
          end
        end
      end
    end
    return out_str
  end
  private :get_proj_low

  def accum_proj_low(mu,y)
    low_comp = get_proj_low(mu,y)
    out_str = Array.new(4,nil)
    for i in 1..@@spin
      if (low_comp[i-1])
        out_str[i-1] = "#{@myname}(:,#{i}) = #{@myname}(:,#{i}) " + low_comp[i-1]
      end
    end
    return out_str.compact
  end

  def assign_proj_low(mu,y)
    low_comp = get_proj_low(mu,y)
    out_str = Array.new(4,nil)
    for i in 1..@@spin
      if (low_comp[i-1])
        out_str[i-1] = "#{@myname}(:,#{i}) = " + low_comp[i-1]
      end
    end
    return out_str.compact
  end

  def assign_mult_u_half(u,y,sign)
    str = Array.new((@@spin*@@col)/2,nil)
    if (sign > 0)
      for j in 1..@@spin/2
      for i in 1..@@col
        str[i-1+@@col*(j-1)] =
             "    #{@myname}(#{i},#{j}) = #{u}(#{i},1) * #{y.get_name}(1,#{j})  &\n" +
             "  &        + #{u}(#{i},2) * #{y.get_name}(2,#{j})  &\n" +
             "  &        + #{u}(#{i},3) * #{y.get_name}(3,#{j})"
      end
      end
    else
      for j in 1..@@spin/2
      for i in 1..@@col
        str[i-1+@@col*(j-1)] =
             "    #{@myname}(#{i},#{j}) = conjg(#{u}(1,#{i})) * #{y.get_name}(1,#{j})  &\n" +
             "  &        + conjg(#{u}(2,#{i})) * #{y.get_name}(2,#{j})  &\n" +
             "  &        + conjg(#{u}(3,#{i})) * #{y.get_name}(3,#{j})"
      end
      end
    end
    return str.compact
  end

  def assign_mult_u_full(u,y,sign)
    str = Array.new((@@spin*@@col)/2,nil)
    if (sign > 0)
      for j in 1..@@spin
      for i in 1..@@col
        str[i-1+@@col*(j-1)] =
             "    #{@myname}(#{i},#{j}) = #{u}(#{i},1) * #{y.get_name}(1,#{j})  &\n" +
             "  &        + #{u}(#{i},2) * #{y.get_name}(2,#{j})  &\n" +
             "  &        + #{u}(#{i},3) * #{y.get_name}(3,#{j})"
      end
      end
    else
      for j in 1..@@spin
      for i in 1..@@col
        str[i-1+@@col*(j-1)] =
             "    #{@myname}(#{i},#{j}) = conjg(#{u}(1,#{i})) * #{y.get_name}(1,#{j})  &\n" +
             "  &        + conjg(#{u}(2,#{i})) * #{y.get_name}(2,#{j})  &\n" +
             "  &        + conjg(#{u}(3,#{i})) * #{y.get_name}(3,#{j})"
      end
      end
    end
    return str.compact
  end

end

class FieldDWSpinor

  @@site = ["","ix","iy","iz","it","iw"]

  def initialize(myname)
    @myname = myname
  end

  def get_name
    return @myname
  end

  def site_shift(i=0,j=0)
    str = @@site.clone
    if (i > 0) 
      str[i] = str[i]+"up"
    else
      str[-i] = str[-i]+"dn"
    end
    if (j > 0) 
      str[j] = str[j]+"up"
    else
      str[-j] = str[-j]+"dn"
    end
    for i in 1..5
      str[i].gsub!(/updn|dnup/,"")
    end
    return str[5]+","+str[4]+","+str[3]+","+str[2]+","+str[1]
  end
  private :site_shift

  def field
    str = @myname + "%s(" + site_shift(0,0) + ")%y"
    return Spinor.new(str)
  end

  def field_shift(i=0,j=0)
    str = @myname + "%s(" + site_shift(i,j) + ")%y"
    return Spinor.new(str)
  end
end 

class FieldGauge
  @@site = ["","ix","iy","iz","it","iw"]

  def initialize(myname,type)
    @myname = myname
    @mu   = 1
    @munu = 1
    @type = type
  end

  def get_name
    return @myname
  end

  def get_type
    return @type
  end

  def site_shift(i=0,j=0)
    str = @@site.clone
    if (i > 0) 
      str[i] = str[i]+"up"
    else
      str[-i] = str[-i]+"dn"
    end
    if (j > 0) 
      str[j] = str[j]+"up"
    else
      str[-j] = str[-j]+"dn"
    end
    for i in 1..5
      str[i].gsub!(/updn|dnup/,"")
    end
    return str[4]+","+str[3]+","+str[2]+","+str[1]
  end
  private :site_shift

  def dir(*mu)
    @mu = mu[0]
    if (@type.downcase == "tensor")
      if ( mu.length != 2)
        print "#{@myname} is tensor type. please supply two index for direction.\n "
        abort
      else
        @munu = mu[1]
      end
    end
    return self
  end

  def field
    if (@type.downcase == "vector") 
      str = @myname + "%s(" + @mu.to_s + "," + site_shift(0,0) + ")%u"
      return str
    end
    if (@type.downcase == "tensor")
      str = @myname + "%s(" + @mu.to_s + "," + @munu.to_s + "," + site_shift(0,0) + ")%u"
      return str
    end
  end

  def field_shift(i=0,j=0)
    if (@type.downcase == "vector") 
      str = @myname + "%s(" + @mu.to_s + "," + site_shift(i,j) + ")%u"
      return str
    end
    if (@type.downcase == "tensor")
      str = @myname + "%s(" + @mu.to_s + "," + @munu.to_s + "," + site_shift(i,j) + ")%u"
      return str
    end
  end
end

class SiteRestriction

  def initialize(mode,ordering="normal")
    @mode = mode.downcase
    @ordering = ordering.downcase
    @if_depth = 0
  end

  def set_mode(mode)
    @mode = mode.downcase
    @if_depth = 0
  end

  def set_ordering(ordering)
    @ordering = ordering.downcase
    @if_depth = 0
  end

  def get_mode
    return @mode
  end

  def diff_site_index_blk(i,j)
    strb    = ["","ibx","iby","ibz"]
    strg    = ["","igx","igy","igz"]
    str     = ["","","","","it","iw"]

    strbup  = ["","ibxup","ibyup","ibzup"]
    strgup  = ["","igxup","igyup","igzup"]
    strup   = ["","","","","itup","iwup"]

    strbdn  = ["","ibxdn","ibydn","ibzdn"]
    strgdn  = ["","igxdn","igydn","igzdn"]
    strdn   = ["","","","","itdn","iwdn"]

    strbN   = ["","NT*NS*NGX*NGY*NGZ*NBY*NBZ","NT*NS*NGX*NGY*NGZ*NBZ","NT*NS*NGX*NGY*NGZ"]
    strgN   = ["","NT*NS*NGY*NGZ","NT*NS*NGZ","NT*NS"]
    strN    = ["",       "",   "", "","NS","1"]

    str3 = ""
    if (3 < i)
      str3 = str3       +  strN[i] + "*(" + strup[i] + "-" + str[i] + ")"
    elsif (0 < i && i < 4)
      str3 = str3       + strgN[i] + "*(" + strgup[i] + "-" + strg[i] + ") + " +
                          strbN[i] + "*(" + strbup[i] + "-" + strb[i] + ")"

    elsif (-4 < i && i < 0)
      str3 = str3       + strgN[-i] + "*(" + strgdn[-i] + "-" + strg[-i] + ") + " +
                          strbN[-i] + "*(" + strbdn[-i] + "-" + strb[-i] + ")"
    elsif (i < -3)
      str3 = str3       +  strN[-i] + "*(" + strdn[-i] + "-" + str[-i] + ")"
    end

    str3 = str3 + " &\n  & "

    if (3 < j)
      str3 = str3 + " + " +  strN[j] + "*(" + strup[j] + "-" + str[j] + ")"
    elsif (0 < j && j < 4)
      str3 = str3 + " + " + strgN[j] + "*(" + strgup[j] + "-" + strg[j] + ") + " +
                            strbN[j] + "*(" + strbup[j] + "-" + strb[j] + ")"
    elsif (-4 < j && j < 0)
      str3 = str3 + " + " + strgN[-j] + "*(" + strgdn[-j] + "-" + strg[-j] + ") + " +
                            strbN[-j] + "*(" + strbdn[-j] + "-" + strb[-j] + ")"
    elsif (j < -3)
      str3 = str3 + " + " +  strN[-j] + "*(" + strdn[-j] + "-" + str[-j] + ")"
    end
    return "(" + str3 + ")"
  end

  def diff_site_index(i,j)
    str    = ["","ix","iy","iz","it","iw"]
    strup  = ["","ixup","iyup","izup","itup","iwup"]
    strdn  = ["","ixdn","iydn","izdn","itdn","iwdn"]
    str2 = ["","NS*NT*NZ*NY","NS*NT*NZ","NS*NT","NS","1"]
    str3 = "("
    if (i > 0)
      str3 = str3 + " + " + str2[i] + "*(" + strup[i] + "-" + str[i] + ")"
    elsif (i < 0)
      str3 = str3 + " + " + str2[-i] + "*(" + strdn[-i] + "-" + str[-i] + ")"
    end
    if (j > 0)
      str3 = str3 + " + " + str2[j] + "*(" + strup[j] + "-" + str[j] + ")"
    elsif (j < 0)
      str3 = str3 + " + " + str2[-j] + "*(" + strdn[-j] + "-" + str[-j] + ")"
    end
    str3 = str3 + ")"
    return str3.sub(/\( \+ /,"(")
  end
  private :diff_site_index

  def extra_boundary(i)
    if (i == +5)
      puts "    kac = (iw/NS)"
      puts "    fac = (1-kac)-mq*kac"
    end
    if (i == -5)
      puts "    kac = ((NS-iw+1)/NS)"
      puts "    fac = (1-kac)-mq*kac"
    end
  end

  def put_if(i=0,j=0)
    if (@mode == "forw")
      if (@ordering == "normal")
        puts "  if (" + diff_site_index(i,j) + " > 0) then"
      elsif (@ordering == "blocknormal")
        puts "  if (" + diff_site_index_blk(i,j) + " > 0) then"
        if (i.abs <4)
          puts "!$OMP BARRIER"
        end
      end
    elsif (@mode == "back")
      if (@ordering == "normal")
        puts "  if (" + diff_site_index(i,j) + " < 0) then"
      elsif (@ordering == "blocknormal")
        puts "  if (" + diff_site_index_blk(i,j) + " < 0) then"
        if (i.abs <4)
          puts "!$OMP BARRIER"
        end
      end
    elsif (@mode == "all")
      puts ""
    else
      print "error in mode\n"
      abort
    end
    @if_depth = @if_depth + 1 
  end

  def put_endif
    if (@mode != "all")
      puts "  endif"
    end
    @if_depth = @if_depth - 1 
  end

end
