#!/usr/bin/ruby
#
# Generates SSOR Kernel for Hermitian Squared Domainwall operator (D'D)
#
# $Id: GenMultSSOR.rb,v 1.3 2010/09/10 06:25:10 ishikawa Exp $
#

require 'FieldClasses'

state = SiteRestriction.new("","blocknormal")

amu = Array.new(5)
for i in 1..5
   amu[i-1] = i
end

puts "!================================================="

yt = Spinor.new("yt%y")
st = Spinor.new("st%y")
y  = Spinor.new("y")
w  = Spinor.new("w")

z  = FieldDWSpinor.new("z")
b  = FieldDWSpinor.new("q")
x  = FieldDWSpinor.new("dq")
vu = FieldGauge.new("vu","vector")

##############
# st = Hop z
##############
puts "    st%y(:,:) = Z0"

##########################################
#  st = D'z
##########################################
for mu in amu.reverse
  puts  "!---------------------------"
  print "!  mu=",mu,"\n"
  puts  "!---------------------------"

  ##########################################
  #  st = st + (1+gamma_mu)U_mu(n) z(n+mu)
  ##########################################
  state.extra_boundary(+mu)
  if (mu != 5)
    puts  y.assign_proj_up(+mu,z.field_shift(+mu))
    puts  w.assign_mult_u_half(vu.dir(mu).field,y,+1)
    puts st.accum_proj_low(+mu,w)
  else
    puts  y.assign_proj_up(+mu,z.field_shift(+mu))
    puts st.accum_proj_low(+mu,y)
  end

  ##############################################
  #  st = st + (1-gamma_mu)U_mu(n-mu)^+ z(n-mu)
  ##############################################
  state.extra_boundary(-mu)
  if (mu != 5)
    puts  y.assign_proj_up(-mu,z.field_shift(-mu))
    puts  w.assign_mult_u_half(vu.dir(mu).field_shift(-mu),y,-1)
    puts st.accum_proj_low(-mu,w)
  else
    puts  y.assign_proj_up(-mu,z.field_shift(-mu))
    puts st.accum_proj_low(-mu,y)
  end

end

######################################
# st = (z - kappa * st) = D'D x = D'z
######################################
puts ""
puts "    st%y(:,:) = " + z.field.get_name + "(:,:) - kappa * st%y(:,:)"
puts ""

#################################################
# multiply inverse diagnal part of (D'D)
#  st = Diag \ st
#################################################
puts "    kac  = 1 - (((NS-iw+1)/NS) + (iw/NS))"
puts "    fac  = (1.0_DP + 16*kappa2 + 2*kappa2*(1.0_DP + mq**2))*(1-kac) + (1.0_DP + 20*kappa2)*kac"
puts "    kac  = ((NS-iw+1)/NS) - (iw/NS)"
puts "    fac2 = 2*kappa2*(1.0_DP - mq**2)*kac"

#####################
# yt = gamma_5 st
#####################
puts ""
puts yt.assign_mult_gamma(5,st)

puts ""
puts "    st%y(:,:) = (st%y(:,:)*fac - yt%y(:,:)*fac2)/(fac**2-fac2**2)"
puts ""

#################################################
# overrelaxation
# st = b - st*omega
#################################################
puts "    st%y(:,:) =  " + b.field.get_name + "(:,:) - st%y(:,:)*omega"
puts ""

##################################################
# Update solution
# x(n) = 0 + st(n)
##################################################
puts "if ((ibx == 1 .or. ibx == NBX) .or. &"
puts "    (iby == 1 .or. iby == NBY) .or. &"
puts "    (ibz == 1 .or. ibz == NBZ)) then"
puts "!$OMP BARRIER"
puts "endif"

#puts "    " + x.field.get_name + "(:,:) = " + x.field.get_name + "(:,:) + st%y(:,:)"
puts "    " + x.field.get_name + "(:,:) = st%y(:,:)"

##################################################
# Update auxiliary vector z
# z = D x (keep this relation)
##################################################

#######################
# z(n) = z(n) + st(n)
#######################

puts "    " + z.field.get_name + "(:,:) = " + z.field.get_name + "(:,:) + st%y(:,:)"

puts ""
puts "    st%y(:,:) = -kappa*st%y(:,:)"

for mu in amu.reverse

  ###############################################################
  # z(n-mu) = z(n-mu) + (1-gamma_mu)U_mu(n-mu) st(n)
  ###############################################################
  puts ""
  state.extra_boundary(-mu)
  if (mu != 5)
    puts y.assign_proj_up(-mu,st)
    puts w.assign_mult_u_half(vu.dir(mu).field_shift(-mu),y,+1)
    puts z.field_shift(-mu).accum_proj_low(-mu,w)
  else
    puts y.assign_proj_up(-mu,st)
    puts z.field_shift(-mu).accum_proj_low(-mu,y)
  end


  ################################################################
  # z(n+mu) = z(n+mu) - kappa * (1+gamma_mu)U_mu(n)^dag st(n)
  ################################################################
  puts ""
  state.extra_boundary(+mu)
  if (mu != 5)
    puts y.assign_proj_up(+mu,st)
    puts w.assign_mult_u_half(vu.dir(mu).field,y,-1)
    puts z.field_shift(+mu).accum_proj_low(+mu,w)
  else
    puts y.assign_proj_up(+mu,st)
    puts z.field_shift(+mu).accum_proj_low(+mu,y)
  end

end
