!-------------------------------------------------------------------------------
!= PHMC_HMC Configuration file, Fortran 9X, Free Format
!
!== Version
!  $Id: config.h 61 2012-02-16 14:34:58Z ishikawa $
!-------------------------------------------------------------------------------
#ifndef LQFT_CONFIG_H_
#define LQFT_CONFIG_H_


!-------------------------------------------
! VERSION
!-------------------------------------------
#define _VERSION_ "refactored-v1.597-SF0.945"

!-------------------------------------------
! Section: Set Lattice size
!-------------------------------------------
#define _NTX  16
#define _NTY  16
#define _NTZ  16
#define _NTT  64
!--------------
! End Section
!--------------

!-------------------------------------------
! Section: Set node size
!-------------------------------------------
#define _NDIMX  1
#define _NDIMY  1
#define _NDIMZ  1
!--------------
! End Section
!--------------

!-------------------------------------------
! Section: Lattice boundary condtion
!-------------------------------------------
!===========================================================================
! Fermion and Gauge
!
! Schroedinger Functional Boundary setups
! 
! If _SF is defined, the SF boundary condition is employed.
! _SF_BOUNDARY_TYPE == 0, phi=pi*( -1, 0, 1)/3, phi_prime=pi*( -3, 1, 2)/3
! _SF_BOUNDARY_TYPE == 1, phi=pi*( -1, 0, 1)/6, phi_prime=pi*( -5, 2, 3)/6
!---------------------------------------------------------------------------
!#define _SF
#undef _SF
!#define _SF_BOUNDARY_TYPE 0
#define _SF_BOUNDARY_TYPE 1

!---------------------------------------------------------------------------
! Fermion boundary condition
! temporal direction only
!  PBC : periodic b.c.
! APBC : anti-periodic b.c.
!---------------------------------------------------------------------------
#ifndef _SF
!#define _FERMI_BOUNDARY_TYPE_T  PBC
#define _FERMI_BOUNDARY_TYPE_T  APBC
#endif

!--------------
! End Section
!--------------

!-------------------------------------------
! Action configuration
!-------------------------------------------
!----------------------------
! Section: Gauge action
!----------------------------
!-------------------------------------------
! If _RGGAUGE is defined, we use gauge 
!     action including rectangle loops.
!-------------------------------------------
!#define _RGGAUGE
#undef _RGGAUGE
!--------------
! End Section
!--------------

!-------------------------------------------
! Section: Stout smearing for Quark actions
!-------------------------------------------
!---------------------------------------------------
! define _STOUT to use stout smeared quark actions
! all quark actions are stout smeared with 
! an identical smearing parameter set.
!---------------------------------------------------
!#define _STOUT
#undef _STOUT
!--------------
! End Section
!--------------

!------------------------------------
! Section: Wilson type quark action
!------------------------------------
!-------------------------------------------
! If _CLOVER is defined, we use 
! the O(a)-improved Wilson (clover) action.
!-------------------------------------------
!#define _CLOVER
#undef _CLOVER

!-------------------------------------------
! choose hmc Wilson-Dirac matrix inverse solver
! bicgstab_alg   : BiCGStab
!-------------------------------------------
#define _HMC_SOLVER_ bicgstab_alg

!-------------------------------------------
! choose clover term matrix inverse solver
! clvinv_lu  : use LU decomposition
! clvinv_ldl : use LDL^+ decomposition
!-------------------------------------------
!#define _CLV_SOLVER  clvinv_lu
#define _CLV_SOLVER  clvinv_ldl

!--------------
! End Section
!--------------

!------------------------------------
! Section: Block HMC solver type
!------------------------------------
!-------------------------------------------
! if  _USE_FORCE_CG_SOLVER is defined
! and GPU is used Hermitan CG solver is used
! in the force computation.
!-------------------------------------------
!#define _USE_FORCE_CG_SOLVER
!#define _USE_HAMIL_CG_SOLVER
#undef _USE_FORCE_CG_SOLVER
#undef _USE_HAMIL_CG_SOLVER

!------------------------------------
! Section: Measurement solver
!------------------------------------
!-------------------------------------------
! measurement wilson/clover solver
!  BLOCK solver for singele NODE only
!-------------------------------------------
!#define _USE_BLOCK_SOLVER
#undef _USE_BLOCK_SOLVER
!#define _USE_MEASURE_CG_SOLVER
#undef _USE_MEASURE_CG_SOLVER
!--------------
! End Section
!--------------

!-------------------------------------------
! do not touch the following lines
!-------------------------------------------
#if _NDIMX > 1
#define _DEPTHX 1
#else
#define _DEPTHX 0
#endif
#if _NDIMY > 1
#define _DEPTHY 1
#else
#define _DEPTHY 0
#endif
#if _NDIMZ > 1
#define _DEPTHZ 1
#else
#define _DEPTHZ 0
#endif
#undef _singlePU
#if (_DEPTHX==0) && (_DEPTHY==0) && (_DEPTHZ==0)
#define _singlePU
#endif

!-------------------------------------------
! Section: Set node Lattice size
!-------------------------------------------
#define _NX (_NTX/_NDIMX)
#define _NY (_NTY/_NDIMY)
#define _NZ (_NTZ/_NDIMZ)
!--------------
! End Section
!--------------

!-------------------------------------------
! Section: check block solver
!-------------------------------------------
#ifdef _USE_BLOCK_SOLVER
#ifndef _singlePU
#error Block-solver cannot be used in node-parallel lattice setup. (undef _USE_BLOCK_SOLVER)
#endif
#endif
!--------------
! End Section
!--------------

!--------------------------------
! Section: boundary condition
!--------------------------------
#ifdef _SF
#  undef _FERMI_BOUNDARY_TYPE_T
#else
#  define  PBC  1
#  define APBC -1
#endif
!--------------
! End Section
!--------------

!--------------------------------
! Section: check measurment solver
!--------------------------------
#ifndef _USE_BLOCK_SOLVER
#ifdef _USE_MEASURE_CG_SOLVER
#error CG solver for measurement is available only with block solver. Undefine USE_MEAREUE_CG_SOLVER
#endif
#endif
!--------------
! End Section
!--------------

!!============================================
!! Macros below are used for testing program
!!
!-------------------------------------------
! PARALLEL CHECK 
! (do not define for productive runs!)
!-------------------------------------------
!#define _PARALLELCHECK

!-------------------------------------------
! Reversiblity check
! (do not define for productive runs!)
!-------------------------------------------
!#define _REVCHECK
#undef _REVCHECK

!-------------------------------------------
! Set do loop coalescing depth
! OpenMP Collapse clause
!-------------------------------------------
#undef _OMPCLPS
!#define _OMPCLPS 1
!#define _OMPCLPS 2
!#define _OMPCLPS 3


#if defined(_OMPCLPS)
#define  _CLPS_   COLLAPSE(_OMPCLPS) 
#define  _CLPS2_  COLLAPSE(2)
#define  _CLPS3_  COLLAPSE(3)
#else
#define  _CLPS_ 
#define  _CLPS2_
#define  _CLPS3_
#endif

!============================================
! Macros below are obsolete.
!

#endif
